package main

import (
	"bufio"
	"bytes"
	"errors"
	"flag"
	"fmt"
	"log"
	"os"
	"os/exec"
	"os/user"
	"path/filepath"
	"strconv"
	"strings"
	"sync"

	"git.autistici.org/ale/liber"
)

var (
	databaseDir = flag.String("db-dir", "~/.liber", "database directory")
	bookDir     = flag.String("book-dir", "", "books directory")
	update      = flag.Bool("update", false, "update the db")
	search      = flag.Bool("search", false, "search something")
	remotesync  = flag.String("sync", "", "push data to remote server")
	httpserver  = flag.String("http-server", "", "start the HTTP server on the specified address")
)

// Various ways to ask a user to choose something.

// Prompt user using stdin. Kind of annoying because it interferes
// with logging on stderr. It is used as a fallback.
func promptUserStdin(path string, choices []*liber.Metadata) (*liber.Metadata, error) {
	fmt.Printf("\n[*] Possible matches for %s:\n\n", path)
	for idx, md := range choices {
		fmt.Printf("    %d) %s\n", idx+1, md.String())
	}
	prompt := "Pick one, or Enter to skip: "

	rdr := bufio.NewReader(os.Stdin)
	for {
		fmt.Printf(prompt)
		os.Stdout.Sync()
		result, err := rdr.ReadString('\n')
		if err != nil || result == "" {
			break
		}
		idx, err := strconv.Atoi(strings.TrimSpace(result))
		if err != nil {
			fmt.Printf("%v\n", err)
			continue
		}
		if idx < 1 || idx > len(choices) {
			fmt.Printf("Insert a number between 1 and %d.\n", len(choices))
			continue
		}
		return choices[idx-1], nil
	}
	return nil, nil
}

func findProgram(progs []string) (string, error) {
	for _, p := range progs {
		if path, err := exec.LookPath(p); err == nil {
			return path, nil
		}
	}
	return "", errors.New("not found")
}

var (
	dialogProg     string
	dialogProgInit bool
)

func getDialogProg() (string, error) {
	if !dialogProgInit {
		dialogProgInit = true
		dialogProgs := []string{"whiptail", "dialog"}
		if os.Getenv("DISPLAY") != "" {
			dialogProgs = append([]string{"gdialog", "xdialog"}, dialogProgs...)
		}
		if p, err := findProgram(dialogProgs); err == nil {
			dialogProg = p
		}
	}
	if dialogProg == "" {
		return "", errors.New("not found")
	}
	return dialogProg, nil
}

// Prompt user using 'dialog', or a graphical variant if X11 is detected.
func promptUserDialog(path string, choices []*liber.Metadata) (*liber.Metadata, error) {
	dialog, err := getDialogProg()
	if err != nil {
		return nil, err
	}

	args := []string{
		"--title", "Metadata Chooser",
		"--menu", fmt.Sprintf("Possible matches for %s:", path),
		"0", "0", "0",
	}
	for idx, md := range choices {
		args = append(args, strconv.Itoa(idx+1))
		args = append(args, md.String())
	}
	log.Printf("running: %s %v", dialog, args)
	var output bytes.Buffer
	cmd := exec.Command(dialog, args...)
	cmd.Stdin = os.Stdin
	cmd.Stdout = os.Stdout
	cmd.Stderr = &output
	if err := cmd.Run(); err != nil {
		log.Printf("dialog failed: %v", err)
		return nil, err
	}
	result, err := strconv.Atoi(strings.TrimSpace(output.String()))
	if err != nil {
		return nil, nil
	}
	return choices[result-1], nil
}

var promptMutex sync.Mutex

func promptUser(path string, choices []*liber.Metadata) *liber.Metadata {
	promptMutex.Lock()
	defer promptMutex.Unlock()

	result, err := promptUserDialog(path, choices)
	if err != nil {
		result, err = promptUserStdin(path, choices)
		if err != nil {
			return nil
		}
	}
	return result
}

func doUpdate(db *liber.Database, dir string) {
	db.Update(dir, promptUser)
}

func doSync(db *liber.Database, remoteAddr string) {
	sc := liber.NewRemoteServer(remoteAddr)
	if err := db.Sync(sc); err != nil {
		log.Fatal(err)
	}
}

func doSearch(db *liber.Database, query string) {
	results, err := db.Search(query, 0, 100)
	if err != nil {
		log.Fatal(err)
	}
	if results.NumResults == 0 {
		fmt.Printf("No results.\n")
	} else {
		fmt.Printf("%d results found:\n\n", results.NumResults)
		for i, r := range results.Results {
			fmt.Printf("%d) %s\n", i+1, r.Metadata.String())
			fmt.Printf("      %s\n", r.Path)
		}
	}
}

func doHttpServer(db *liber.Database, addr string) {
	storage := liber.NewFileStorage(expandTilde(*bookDir), 2)
	cache := liber.NewFileStorage(filepath.Join(expandTilde(*databaseDir), "cache"), 2)
	server := liber.NewHttpServer(db, storage, cache, addr)
	log.Fatal(server.ListenAndServe())
}

func b2i(b bool) int {
	if b {
		return 1
	}
	return 0
}

func expandTilde(path string) string {
	if path[:2] == "~/" {
		curUser, _ := user.Current()
		return filepath.Join(curUser.HomeDir, path[2:len(path)])
	}
	return path
}

func main() {
	log.SetFlags(0)
	flag.Parse()
	nset := b2i(*update) + b2i(*search) + b2i(*httpserver != "") + b2i(*remotesync != "")
	if nset != 1 {
		log.Fatal("Must specify one of --update, --sync, --search or --http-server")
	}
	if *update && *bookDir == "" {
		log.Fatal("Must specify --book-dir with --update")
	}

	dbdir := expandTilde(*databaseDir)
	db, err := liber.NewDb(dbdir)
	if err != nil {
		log.Fatal(err)
	}
	defer db.Close()

	if *update {
		// Redirect logging to dbdir/update.log.
		logf, err := os.OpenFile(filepath.Join(dbdir, "update.log"), os.O_WRONLY|os.O_APPEND|os.O_CREATE, 0644)
		if err == nil {
			defer logf.Close()
			log.SetOutput(logf)
		}

		doUpdate(db, expandTilde(*bookDir))
	} else if *remotesync != "" {
		doSync(db, *remotesync)
	} else if *search {
		query := strings.Join(flag.Args(), " ")
		if query == "" {
			log.Fatal("No query specified")
		}
		doSearch(db, query)
	} else if *httpserver != "" {
		doHttpServer(db, *httpserver)
	}
}
