package liber

import (
	"io/ioutil"
	"os"
	"testing"
)

var testdbdir = "/tmp/liber-test-database"

type testDatabase struct {
	db   *Database
	path string
}

func (td *testDatabase) Close() {
	td.db.Close()
	os.RemoveAll(td.path)
}

func newTestDatabase(t *testing.T) (*testDatabase, *Database) {
	path, _ := ioutil.TempDir("", "testdb-")
	db, err := NewDb(path)
	if err != nil {
		t.Fatalf("NewDb(): %v", err)
	}

	book := testEbook()
	if err = db.PutBook(book); err != nil {
		t.Fatalf("PutBook(): %v", err)
	}

	return &testDatabase{db: db, path: path}, db
}

func testEbook() *Book {
	return &Book{
		Id:       NewID(),
		Path:     "/path/to/ebook",
		FileType: ".epub",
		Metadata: &Metadata{
			Title:   "20,000 Leagues under the sea",
			Creator: []string{"Jules Verne"},
			ISBN:    []string{"1234"},
		},
	}
}

func TestDatabase_Search(t *testing.T) {
	td, db := newTestDatabase(t)
	defer td.Close()

	r, err := db.Search("jules verne", 0, 100)
	if err != nil {
		t.Fatal(err)
	}
	if r.NumResults != 1 {
		t.Fatalf("NumResults != 1 - %#v", r)
	}
	if len(r.Results) != 1 {
		t.Fatalf("len(Results) != 1 - %#v", r)
	}

	r, err = db.Search("italo calvino", 0, 100)
	if err != nil {
		t.Fatal(err)
	}
	if r.NumResults > 0 {
		t.Fatalf("unexpected results: %v", r)
	}
}

func TestDatabase_Find(t *testing.T) {
	td, db := newTestDatabase(t)
	defer td.Close()

	// Find using ISBN.
	book := testEbook()
	if _, err := db.Find(book.Metadata); err != nil {
		t.Fatal("With ISBN: ", err)
	}

	// Find using title/author.
	book.Metadata.ISBN = nil
	if _, err := db.Find(book.Metadata); err != nil {
		t.Fatal("With title/author: ", err)
	}

	// Find only using title (should fail).
	book.Metadata.Creator = nil
	if result, err := db.Find(book.Metadata); err == nil {
		t.Fatalf("Title only: no error, result = %v", result)
	}
}

func TestDatabase_Autocomplete(t *testing.T) {
	td, db := newTestDatabase(t)
	defer td.Close()

	r, err := db.Autocomplete("jul")
	if err != nil {
		t.Fatal(err)
	}
	if r.NumResults == 0 {
		t.Error("No results")
	}

	r, err = db.Autocomplete("foo")
	if err != nil {
		t.Fatal(err)
	}
	if r.NumResults != 0 {
		t.Errorf("Unexpected results: %v", r)
	}
}
