package liber

import (
	"errors"
	"log"
	"os"
	"sync"

	"git.autistici.org/ale/liber/util"
)

const (
	SourceDB = 1 << iota
	SourceFS
)

type MetadataChooserFunc func(string, []*Metadata) *Metadata

type fileData struct {
	source int
	path   string
	id     BookId
	info   os.FileInfo
}

func (f fileData) toLiberFile(haserr bool) *File {
	return &File{
		Path:  f.path,
		Mtime: f.info.ModTime(),
		Size:  f.info.Size(),
		Id:    f.id,
		Error: haserr,
	}
}

func dbFileScanner(db *Database, fileCh chan fileData) {
	for iter := db.Scan(FileBucket); iter.Valid(); iter.Next() {
		var f File
		if err := iter.Value(&f); err != nil {
			continue
		}
		fileCh <- fileData{
			source: SourceDB,
			path:   f.Path,
			id:     f.Id,
		}
	}
}

func localFileScanner(db *Database, basedir string, fileCh chan fileData) {
	util.NewDefaultWalker().Walk(basedir, func(path string, info os.FileInfo, err error) error {
		if err != nil {
			return nil
		}

		fileCh <- fileData{
			source: SourceFS,
			path:   path,
			info:   info,
		}
		return nil
	})
}

func differ(db *Database, basedir string) chan fileData {
	fileCh := make(chan fileData, 100)
	outCh := make(chan fileData, 100)
	var wg sync.WaitGroup
	wg.Add(2)

	// Start two sources in parallel and send their output to fileCh.
	go func() {
		localFileScanner(db, basedir, fileCh)
		wg.Done()
	}()
	go func() {
		dbFileScanner(db, fileCh)
		wg.Done()
	}()
	// Once they are done, close the channel.
	go func() {
		wg.Wait()
		close(fileCh)
	}()
	go func() {
		// Detect files that have not changed, i.e. appear in
		// the database and the filesystem. Keep track of book
		// IDs so that once all entries have been processed we
		// can delete those books from the database where the
		// original file has been removed.
		allSources := SourceDB | SourceFS
		tmp := make(map[string]int)
		ids := make(map[string]BookId)
		for f := range fileCh {
			// log.Printf("differ: %#v", f)
			tmp[f.path] |= f.source
			// Delete entries as soon as we've seen them
			// from both sources.
			if tmp[f.path] == allSources {
				// log.Printf("differ: dropping %s", f.path)
				delete(tmp, f.path)
				delete(ids, f.path)
			}
			if f.source == SourceFS {
				outCh <- f
			} else {
				ids[f.path] = f.id
			}
		}
		for path, value := range tmp {
			if value == SourceDB {
				log.Printf("removing book %s", path)
				db.DeleteBook(ids[path])
			}
		}
		close(outCh)
	}()
	return outCh
}

func adder(db *Database, chooser MetadataChooserFunc, fileCh chan fileData) {
	for f := range fileCh {
		var oldid BookId
		if oldfile, err := db.GetFile(f.path); err == nil {
			if !oldfile.HasChanged(f.info) {
				continue
			}
			oldid = oldfile.Id
		}

		var err error
		f.id, err = importBook(db, f, oldid, chooser)
		if err != nil {
			log.Printf("Could not add %s: %v", f.path, err)
			continue
		}

		file := f.toLiberFile(err != nil)
		if err := db.PutFile(file); err != nil {
			log.Println(err)
		}
	}
}

func importBook(db *Database, f fileData, oldid BookId, chooser MetadataChooserFunc) (BookId, error) {
	// Attempt direct metadata extraction.
	book, err := Parse(f.path)
	if err != nil {
		return 0, err
	}

	if oldid != 0 {
		book.Id = oldid
	} else {
		book.Id = NewID()
	}

	// Check if a Calibre OPF file exists.
	if opfmeta, err := opfOpen(opfMetadataPath(f.path)); err == nil {
		book.Metadata.Merge(opfmeta)
	} else {
		// No local metadata, use Google Books to retrieve
		// more information on the book. Ask the user to
		// choose in case there are multiple results.
		candidates, err := LookupGoogleBooks(book.Metadata)
		if err == nil && len(candidates) > 0 {
			if len(candidates) == 1 {
				log.Printf("found Google Books match: %s", candidates[0].String())
				book.Metadata.Merge(candidates[0])
			} else {
				if userchoice := chooser(f.path, candidates); userchoice != nil {
					book.Metadata.Merge(userchoice)
				}
			}
		}
	}

	// Check if the book metadata looks ok.
	if !book.Metadata.Sufficient() {
		return 0, errors.New("insufficient metadata")
	}

	// Try to find a cover image. Look on the local filesystem
	// first, otherwise check Google Books.
	localCoverPath := opfCoverPath(f.path)
	if _, err := os.Stat(localCoverPath); err == nil {
		book.CoverPath = localCoverPath
	} else if imageData, err := GetGoogleBooksCover(book.Metadata); err == nil {
		imageFileName := book.Path + ".cover.png"
		if imgf, err := os.Create(imageFileName); err != nil {
			log.Printf("Could not save cover image for %d: %v", book.Id, err)
		} else {
			imgf.Write(imageData)
			imgf.Close()
			book.CoverPath = imageFileName
		}
	}

	// Save the book in our database.
	if err := db.PutBook(book); err != nil {
		return book.Id, err
	}
	log.Printf("%s -> %d", f.path, book.Id)
	return book.Id, nil
}

func (db *Database) Update(dir string, chooser MetadataChooserFunc) {
	var wg sync.WaitGroup
	ch := differ(db, dir)
	for i := 0; i < 10; i++ {
		wg.Add(1)
		go func() {
			adder(db, chooser, ch)
			wg.Done()
		}()
	}
	wg.Wait()
}
